<?php

declare(strict_types=1);


namespace App\Arbitration\Inner\ReceiptCalculus;


use App\Arbitration\Inner\ArbitrationChain;
use App\Arbitration\Inner\Receipt\Deal;
use App\Arbitration\Inner\Receipt\Receipt;
use App\Exchanges\MarketInterface;
use App\Utils\Money\Currency;
use App\Utils\Money\Money;
use App\Utils\Order\Order;
use App\Utils\Wallet\WalletInterface;

class Calculator
{

    private WalletInterface $wallet;

    public function __construct(WalletInterface $wallet)
    {
        $this->wallet = $wallet;
    }

    public function calculateArbitrationReceipt(ArbitrationChain $chain): ?Receipt
    {
        $base = $chain->getBase();
        if (!$base) {
            return null;
        }

        $markets = $chain->getMarkets();
        $baseMarkets = array_filter(
            $markets,
            function (MarketInterface $market) use($base): bool
            {
                $pair = $market->getPair();
                return $pair->getBase()->isEqual($base) || $pair->getQuote()->isEqual($base);
            }
        );

        foreach ($baseMarkets as $baseMarket) {
            usort(
                $markets,
                function (MarketInterface $a, MarketInterface $b) use($baseMarket) {
                    if ($a === $baseMarket) {
                        return 1;
                    }

                    if ($b === $baseMarket) {
                        return -1;
                    }

                    return 0;
                }
            );
            $orders = $this->getOrdersAgainstBase($markets, $base);
            $receipt = $this->build($orders, $base);

            if ($receipt) {
                return $receipt;
            }
        }

        return null;
    }

    /**
     * @param MarketInterface[] $markets
     * @param Currency $base
     * @return array
     */
    public function getOrdersAgainstBase(array $markets, Currency $base): array
    {
        $current = $base;
        $result = [];
        while (\count($result) < \count($markets)) {
            foreach ($markets as $market) {
                $pair = $market->getPair();
                if ($pair->getBase()->isEqual($current)) {
                    $orders = $market->getOrderBook()->getBids();
                    $current = $pair->getQuote();
                } elseif ($pair->getQuote()->isEqual($current)) {
                    $orders = $market->getOrderBook()->getBids();
                    $current = $pair->getBase();
                } else {
                    continue;
                }

                $result[] = [
                    'market' => $market,
                    'orders' => $orders,
                    'buy_currency' => $current
                ];

                break;
            }
        }

        return $result;
    }


    protected function build(array $ordersData, Currency $base): ?Receipt
    {
        $available = $this->wallet->getMoneyAmount($base);

        $receiptDeals = [];
        foreach ($ordersData as $orderData) {
            $buyAmount = $this->calculateAmount(
                $ordersData['orders'],
                $ordersData['buy_currency'],
                $available
            );

            if (null === $available) {
                return null;
            }

            /**
             * @var MarketInterface $market
             */
            $market = $ordersData['market'];
            $receiptDeals[] = new Deal(
                $market,
                $available,
                $buyAmount
            );

            $available = new Money(
                bcmul(
                    $buyAmount->getAmount(),
                    (string)(1 - $market->getFee()),
                    $buyAmount->getCurrency()->getPrecision()
                ),
                $buyAmount->getCurrency()
            );
        }
        /**
         * @var Deal $last
         */
        $last = end($receiptDeals);

        if ($this->wallet->getMoneyAmount($base)->compare($last->getAsk()) <= 0) {
            return null;
        }

        return new Receipt(...$receiptDeals);
    }

    protected function calculateAmount(array $orders, Currency $buy, Money $maxAmount): ?Money
    {
        usort(
            $orders,
            function (Order $a, Order $b): int
            {
                return bccomp(
                    bcdiv(
                        $a->getFrom()->getAmount(),
                        $a->getTo()->getAmount(),
                        max($a->getFrom()->getCurrency()->getPrecision(), $a->getTo()->getCurrency()->getPrecision())
                    ),
                    bcdiv(
                        $b->getFrom()->getAmount(),
                        $b->getTo()->getAmount(),
                        max($b->getFrom()->getCurrency()->getPrecision(), $b->getTo()->getCurrency()->getPrecision())
                    ),
                    max(
                        $a->getFrom()->getCurrency()->getPrecision(),
                        $a->getTo()->getCurrency()->getPrecision(),
                        $b->getFrom()->getCurrency()->getPrecision(),
                        $b->getTo()->getCurrency()->getPrecision(),
                    )
                );
            }
        );

        $currency = new Money("0", $buy);
        $precision = max($maxAmount->getCurrency()->getPrecision(), $buy->getPrecision());
        $highestPrice = null;
        foreach ($orders as $order) {
            /**
             * @var Order $order
             */
            $currency->add($order->getTo());
            $highestPrice = new Money(
                bcdiv(
                    $order->getFrom()->getAmount(),
                    $order->getTo()->getAmount(),
                    $precision
                ),
                $precision
            );
            $total = new Money(
                bcmul(
                    $currency->getAmount(),
                    $highestPrice->getAmount(),
                    $precision
                ),
                $order->getFrom()->getCurrency()
            );
            if ($total->compare($maxAmount) <= 0) {
                break;
            }
        }

        if (null === $highestPrice) {
            return null;
        }

        return new Money(
            bcdiv(
                $maxAmount->getAmount(),
                $highestPrice->getAmount(),
                $precision
            ),
            $orders[0]['buy_currency']
        );
    }
}