<?php

declare(strict_types=1);


namespace App\Infrastructure\Arbitration\Inner;


use App\Arbitration\Inner\ArbitrationChain;
use App\Arbitration\Inner\ChainProviderInterface;
use App\Exchanges\ExchangeInterface;
use App\Exchanges\ExchangeProvider;
use App\Exchanges\MarketInterface;
use App\Utils\Money\Currency;
use MongoDB\Client;
use MongoDB\Collection;

class ChainProvider implements ChainProviderInterface
{

    private Collection $collection;

    public function __construct(
        Client $client
    )
    {
        $this->collection = $client->selectCollection('arbitry', 'inner_arbitration_chains');
    }

    /**
     * @inheritDoc
     */
    public function getChainsForExchange(ExchangeInterface $exchange): array
    {
        $rows = $this->collection->find(['exchange' => $exchange->getName()]);

        $markets = $exchange->getMarkets();

        $chains = [];
        foreach ($rows as $row) {
            $chainMarkets = [];
            foreach ($row['chain'] ?? [] as $chainData) {
                foreach ($markets as $market) {
                    if ((string)$market->getPair() === $chainData) {
                        $chainMarkets[] = $market;
                    }
                }
            }

            $base = null;
            if (!empty($row['base'])) {
                $base = new Currency($row['base']['code'], $row['base']['precision']);
            }

            if (!empty($chainMarkets)) {
                $chain = new ArbitrationChain($exchange, ...$chainMarkets);
                if ($base) {
                    $chain->setBase($base);
                }
                $chains[] = $chain;
            }
        }

        return $chains;
    }

    protected function calculateChainHash(ArbitrationChain $chain): string
    {
        return md5(
            $chain->getExchange()->getName()
            . ';'
            . implode(
                ';',
                array_map(
                    static function (MarketInterface $market): string
                    {
                        return (string)$market->getPair();
                    },
                    $chain->getMarkets()
                )
            )
        );
    }

    public function addChain(ArbitrationChain $chain): void
    {
        $base = $chain->getBase();
        $this
            ->collection
            ->updateOne(
                ['_id' => $this->calculateChainHash($chain)],
                [
                    'exchange' => $chain->getExchange()->getName(),
                    'chain' => array_map(
                        static function (MarketInterface $market): string
                        {
                            return (string)$market->getPair();
                        },
                        $chain->getMarkets()
                    ),
                    'base' => $base ? ['code' => $base->getCode(), 'precision' => $base->getPrecision()] : null
                ],
                ['upsert' => true]
            );
    }

    public function removeChain(ArbitrationChain $chain): void
    {
        $this
            ->collection
            ->deleteOne([
                '_id' => $this->calculateChainHash($chain)
            ]);
    }
}